/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MIDI project - MPIA/Leiden
*
* "@(#) $Id: oirShiftMask.c,v 1.5 2009/12/11 10:31:20 jaffe Exp $"
*
* who       when      what
* --------  --------  ----------------------------------------------
* wjaffe    14/06/06  initial version
*/
/***********************************************************************
 *   NAME
 *       oirShiftMask - shift an existing tracking masked based on new photometry images
 *
 *   SYNOPSIS
 *      oirShiftMask -A AFile -B BFile -out outFile -refMask refMaskFile
 *
 *   DESCRIPTION
 *      oirShiftMask cross correlates photometry images against an existing
 *      centers it with respect to the photometric images taken in AOPEN and BOPEN shutter modes
 *      The reference mask is shifted by this amount and outputted
 *
 *   ARGUMENTS
 *     AFile, output of oirChopPhotoImages representing data from
 *              chopped bright star data with shutter setting AOPEN.  
 *
 *     BFile, output of oirChopPhotoImages representing data from  
 *              chopped bright star data with shutter setting BOPEN.  
 *     outFile  name of output FITS table file
 *     refMaskFile, an input mask to be cross correlated with the A/B images.
 *   FILES
 *
 *   ENVIRONMENT
 *       none
 *
 *   COMMANDS
 *        .........
 *
 *   RETURN VALUES
 *       0 if successful
 *
 *----------------------------------------------------------------------*/
/*
 * System Headers
 */
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
/*
 * MIDI NRTS Headers
 */
#include "imaging_fdata.h"
#include "minrtsFftUtil.h"

#define clight 2.997925e14 /* microns per sec */
#define MAXORDER 4 
#define WIDTHORDER 2 
#define SMOOTH 6
#define CUTOFF 0.05
#define EDROP 30

void Usage() {
   printf("%s\n", "Usage:");
   printf("    %s\n\n", "oirShiftMask -A AFile -B BFile -out outFile -refMask refMaskFile ");
   printf("    %s\n", "AFile = AOPEN data reduced by oirChopPhotoImages");
   printf("    %s\n", "BFile = BOPEN data reduced by oirChopPhotoImages");
   printf("    %s\n", "outFile       = name of output file.");
   return;
}

int main (int argc, char** argv) {

   char opt1[12];
   char shutter[12];
   char gris[12];

   char track; 
   char reNorm;

   int iShutter;
   int i,ierr;
   int nx, ny;
   long npix;
   long iPix;
   int  ix, iy, iShift, yMask;

   int maskChan[2];
   
   int iarg;
   char outFileName[130];
   char *refMaskFileName;
   char *inFileNames[2];

   fitsfile *inFile, *inFile2;
   fitsfile *outFile;

   imaging_fdata_table_def *odata, *idata;
   imaging_fdata_row *inData, *outData;
   imaging_detector_table_def *DetectorDef;
   char *history[] = {"HISTORY"};

   int  nChan, iChan;


   float *mask[4];
   float *photData[2][2];
   float *crossCorr[2];
   float *yCoord, *yy, *yyNew, *yData;
   float yShift[2];
   float yShiftMean;
   float *shiftSpline;

   int   maxShift;
   float x,y;

   char *bang="!";
   char *highSens = "HIGH_SENS";
   ierr  = 0;
   maxShift = 6;
   nx    = 0;
   ny    = 0;
   npix  = 0;
   odata = NULL;
   nChan = 0 ;

/*
      ANNOUNCEMENT!
*/
   printf("\n      ******  oirShiftMask ******\n");
/*
      parse input
*/
   outFileName[0]    = 0;
   inFileNames[0]    = NULL;
   inFileNames[1]    = NULL;
   refMaskFileName   = NULL;
   iarg = 1;
   while (iarg < argc ) {
      if (!strcmp("-A", argv[iarg])) inFileNames[0] = argv[++iarg];
      if (!strcmp("-B", argv[iarg])) inFileNames[1] = argv[++iarg];
      if (!strcmp("-out", argv[iarg])) {
         strcpy(outFileName, bang);
         strcat (outFileName, argv[++iarg]);
      }
      if (!strcmp("-refMask", argv[iarg])) refMaskFileName = argv[++iarg];
      ++iarg;
   }
   if (argc <4) {
      Usage();
      return 0;
   }
   if (inFileNames[0]== NULL ) {
      printf("Missing A input file\n");
      Usage();
      return 1;
   }
   if (inFileNames[1]== NULL ) {
      printf("Missing B input file\n");
      Usage();
      return 1;
   }
   if (outFileName[0]==0) {
      printf("Missing output file name\n");
      Usage();
      return 1;
   }
   if (refMaskFileName==NULL) {
      printf("Missing reference Maske File name\n");
      Usage();
      return 1;
   }

/*
      open file Afile and get a lot of parameters 
      also open output file, initialize it and copy header info.
*/
   i = fits_open_file(&inFile, inFileNames[0], READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening file %s\n", ierr, inFileNames[0]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
      shutter, 0, &ierr);
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", opt1, 0, &ierr);
   printf("   beam combiner = %s\n", opt1);
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", gris, 0, &ierr);
   printf("   disperser = %s\n", gris);

   if (strncmp(opt1, highSens, 9) == 0) { 
      nChan = 2; 
      maskChan[0] = 0;
      maskChan[1] = 1;
   } else {
      nChan = 4;
      maskChan[0] = 1;
      maskChan[1] = 2;
   }

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",outFileName, ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
           copy primary header from input to output
*/
   i = fits_copy_header(inFile, outFile, &ierr);
#ifdef HISTORY
/*
           copy HISTORY records from 2nd input
*/
   
   i = fits_open_file(&inFile2, inFileNames[1], READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening file %s\n", ierr, inFileNames[1]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   copy_header_keywords(inFile2, outFile, 1, history, &ierr);
   i = fits_close_file(inFile2, &ierr);
#endif
/*
      move to detector section
*/
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening detector table in %s\n", 
         ierr, inFileNames[0]);
      return 1;
   }
   nx = DetectorDef->naxis[0][0];
   ny = DetectorDef->naxis[0][1];
/*
      allocate accumulator arrays
*/
  
   npix = ny * (long)nx;
   for (iChan=0; iChan<nChan;iChan++) {
      mask[iChan] = (float *)malloc(npix * sizeof(float));
      memset (mask[iChan], 0, npix*sizeof(float));
   }
   for (iChan=0; iChan<2;iChan++) {
         for (iShutter=0;iShutter<2;iShutter++) {
            photData[iShutter][iChan] = (float *)malloc(npix * sizeof(float));
            memset (photData[iShutter][iChan], 0, npix*sizeof(float));
      }
      crossCorr[iChan] = (float *)malloc((2*maxShift+1) * sizeof(float));
      memset (crossCorr[iChan], 0, (2*maxShift+1)*sizeof(float));
   }
   yCoord = (float *)malloc((2*maxShift+1) * sizeof(float));
   for (iy=0;iy<=(2*maxShift);iy++) yCoord[iy] = iy-maxShift;

   i = fits_close_file(inFile, &ierr);
/*
          now read the real data
*/
   for (iShutter=0; iShutter<2; iShutter++) {
      if (iShutter == 0) printf("   Extracting data from AOPEN file:\n");
      if (iShutter == 1) printf("   Extracting data from BOPEN file:\n");
      printf("      %s \n", inFileNames[iShutter]);
     
      ierr = 0;
      i = fits_open_file(&inFile, inFileNames[iShutter], READONLY, &ierr);
      if (ierr != 0) {
         printf("cfitsio error %i opening file %s\n", ierr, inFileNames[iShutter]);
         printf("%s \n",strerror(errno));
         return 1;
      }
      DetectorDef = read_imaging_detector_table(inFile, &ierr);
      if (ierr !=0) {
         printf("error %i opening detector table in %s\n", 
            ierr, inFileNames[iShutter]);
         return 1;
      }
      idata = open_imaging_fdata_table(inFile, &ierr);
      if (ierr !=0) {
         printf("error %i opening imaging data in %s\n",
            ierr, inFileNames[iShutter]);
         return 1;
      }
      inData = make_imaging_fdata_row(idata);
      i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr);
      memcpy(photData[iShutter][0], inData->data[maskChan[0]], npix*sizeof(float));
      memcpy(photData[iShutter][1], inData->data[maskChan[1]], npix*sizeof(float));
      kill_imaging_fdata_row(inData);
   }   /* loop over shutters */
/*
          now read the mask
*/
   ierr = 0;
   printf("   Reading old mask from %s\n", refMaskFileName);
   i = fits_open_file(&inFile, refMaskFileName, READONLY, &ierr);
   if (ierr != 0) {
       printf("cfitsio error %i opening mask file %s\n", ierr, refMaskFileName);
       printf("%s \n",strerror(errno));
       return 1;
    }
    idata = open_imaging_fdata_table(inFile, &ierr);
    if (ierr !=0) {
       printf("error %i opening imaging data in %s\n",
       ierr, inFileNames[iShutter]);
       return;
    }
    inData = make_imaging_fdata_row(idata);
    i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr);
    for (iChan=0;iChan<nChan;iChan++) 
       memcpy(mask[iChan], inData->data[iChan], npix*sizeof(float));
    kill_imaging_fdata_row(inData);
/*
      for A- and B- photometry separately, cross correlate mask and phot data;
      average two interferometric channels together
*/
   for (iShutter=0;iShutter<2;iShutter++) 
      for (iChan=0;iChan<2;iChan++) 
         for (iShift = -maxShift; iShift<maxShift; iShift++) 
            for (iy=0; iy<ny; iy++) {
               yMask = iy - iShift;
               if ((yMask >= 0) && (yMask <ny)) {
                  for (ix=0;ix<nx;ix++) crossCorr[iShutter][iShift+maxShift] +=
                     photData[iShutter][iChan][ix+iy*nx] * mask[maskChan[iChan]][ix+yMask*nx];
            }
            

   }
/*
      find peaks in correlation function
*/
   for (iShutter=0;iShutter <2; iShutter++) yShift[iShutter]= 
      minrtsQPk(2*maxShift-1, yCoord+1, crossCorr[iShutter]+1, &x, 0);
   printf("   For A and B data, shift peaks are %f %f pixels\n",yShift[0], yShift[1]);
   yShiftMean = 0.5*(yShift[0]+yShift[1]);
   printf("   Average peak shift is %f pixels\n",yShiftMean);
/*
      interpolate masks
*/
   yy     = (float *)malloc((ny) * sizeof(float));
   yyNew  = (float *)malloc((ny) * sizeof(float));
   yData  = (float *)malloc((ny) * sizeof(float));
   for (iy=0;iy<ny;iy++) {
      yy[iy]    = (float)iy;
      yyNew[iy] = (float)iy -yShiftMean;
   }
   i = (int)yShiftMean;
   for (iChan=0;iChan<nChan;iChan++) {
      for (ix=0; ix<nx; ix++) {
         for (iy=0;iy<ny;iy++) yData[iy] = mask[iChan][ix+iy*nx];
         shiftSpline = minrtsSpline(ny, yy, yData, ny, yyNew, 0.1);
         for (iy=0;iy<ny;iy++) mask[iChan][ix+iy*nx] = shiftSpline[iy];
         free(shiftSpline);
      }
   }
/*
 *    write position shift info into output header
*/
   i = fits_write_key(outFile, TFLOAT, "YSHIFTA", (void*)(&yShift[0]),
      "BEST SHIFT FROM AOPEN DATA", &ierr);
   i = fits_write_key(outFile, TFLOAT, "YSHIFTB", (void*)(&yShift[1]),
      "BEST SHIFT FROM BOPEN DATA", &ierr);
   i = fits_write_key(outFile, TFLOAT, "YSHIFT", (void*)(&yShiftMean),
      "APPLIED Y-SHIFT", &ierr);
/*
      create output mask 
*/
   odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }
   outData = make_imaging_fdata_row(odata);
   for (iChan=0;iChan<nChan;iChan++) 
      memcpy(outData->data[iChan], mask[iChan], npix*sizeof(float));
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);
/*
        clean up
*/
   i = close_imaging_data_table(outFile, odata, &ierr);
   kill_imaging_detector_table_def(DetectorDef);
   kill_imaging_fdata_row(outData);
   for (iChan=1;iChan<nChan;iChan++) free(mask[iChan]);
   for (iChan=1;iChan<2;iChan++) {
      free(photData[0][iChan]);
      free(photData[1][iChan]);
      free(crossCorr[iChan]);
   }
   free(yCoord);
   free(yy);
   free(yyNew);
   free(yData);
   return 0;
}
